reset()

import numpy
from numpy import linalg
numpy.random.seed(9530) #This seed was maliciosly created

import scipy
from scipy.stats.mstats import mquantiles

import matplotlib
matplotlib.use('SVG')
matplotlib.rc('font',family='M+ 2p',weight='light')
import matplotlib.pyplot as plt
from matplotlib.patches import Ellipse

from sage.misc.decorators import infix_operator
@infix_operator('multiply')
def dot(a,b) : return vector([a[i]*b[i] for i in range(0,min(len(a),len(b)))])

if (numpy.random.randint(10000) != 9501) : print("Warning: Your random number generator did not generate the same number as mine, results may vary")

var('x n')

state_vector = [x];
obs_noise = [n];

def h(state,noise) : return (state[0]+0.5)*(state[0]-0.5)*(state[0]-1.5)/2+noise[0]
h_ev = h(state_vector,obs_noise)
replace_state = [ (str(state_vector[k]),"state["+str(k)+"]") for k in range(0,len(state_vector))]
replace_noise = [ (str(obs_noise[k]),"noise["+str(k)+"]") for k in range(0,len(obs_noise))]

He = matrix([ [diff(h_ev,state_vector[k]) for k in range(0,len(state_vector))] ])
Hee_s = str(He.list())
Hee_s = reduce(lambda u,v : u.replace(v[0],v[1]),replace_state+replace_noise,Hee_s)
Hee_e = 'def Hee(state,noise) : return matrix(' +Hee_s+')'
exec(preparse(Hee_e))

Ve = matrix([ [diff(h_ev,obs_noise[k]) for k in range(0,len(obs_noise[k]))] ])
Vee_s = str(Ve.list())
Vee_s = reduce(lambda u,v : u.replace(v[0],v[1]),replace_state+replace_noise,Vee_s)
Vee_e = 'def Vee(state,noise) : return matrix('+Vee_s+')'
exec(preparse(Vee_e))

#Define the initial state mean
print("The initial state")
sh = vector([0]).n()
P_root = identity_matrix(1)*0.7
P = P_root*P_root.transpose()
print("    Mean")
print(sh.n(digits=3))
print("    Cov")
print(P.n(digits=3))

R_root = identity_matrix(1)*0.05;
R = R_root*R_root.transpose()
vh = vector([0]).n()

load("Generate_Sigma_Points.sage")

SPset = Generate_Order9_nopair(2).n()

num_points = 1000
random_points = [[P_root*vector(numpy.random.normal(0,1,1))+sh,R_root*vector(numpy.random.normal(0,1,1))] for k in range(0,num_points)]
random_x = map(lambda u:u[0][0],random_points)

random_obs = [h(random_points[k][0],random_points[k][1]) for k in range(0,num_points)]

random_mean = numpy.mean(random_obs)
random_xcorr = numpy.cov([random_x,random_obs])
random_xcorr_el = linalg.eig(random_xcorr)
print("Mean:"+str(n(random_mean).n(3))+" Covariance:\n"+str(matrix(random_xcorr).n(3)))

def small_sigma(sn,pn_root,rn,rn_root,spoints) :
    weights = spoints[0]
    ty = [vector([0,0]) for j in range (0,len(weights))]
    spoints_t = spoints.submatrix(1,0).transpose()
    for j in range(0,len(weights)) :
        tsp = spoints_t[j]
        ts = sn + pn_root*vector([tsp[0]])
        tv = rn + rn_root*vector([tsp[1]])
        ty[j] = vector([ts[0],h(ts,tv).n()])
    ym = sum([weights[j]*ty[j] for j in range(0,len(weights))])
    pzz = sum([ weights[j]*(ty[j]-ym).outer_product(ty[j]-ym) for j in range(0,len(weights))])
    pxz = sum([ weights[j]*(pn_root*vector([spoints_t[j][0]])).outer_product(ty[j]-ym) for j in range(0,len(weights))])
    pvz = sum([ weights[j]*(rn_root*vector([spoints_t[j][1]])).outer_product(ty[j]-ym) for j in range(0,len(weights))])
    H = ((pn_root*pn_root.transpose()).inverse()*pxz).transpose()
    V = ((rn_root*rn_root.transpose()).inverse()*pvz).transpose()
    nl = pzz - H*pn_root*pn_root.transpose()*H.transpose()-V*rn_root*rn_root.transpose()*V.transpose()
    pzz = H*pn_root*pn_root.transpose()*H.transpose()+V*rn_root*rn_root.transpose()*V.transpose()+nl
    return (ym,pzz,ty,H,V,nl)

clist = [
    '#ac2b36',
    '#00a250',
    '#00a9e4',
    '#ef5f21'
]

def iterate(sh_in,P_root_in) :
    (ym,pzz,ty,H,V,nl) = small_sigma(sh_in,P_root_in,vector([0]),R_root,SPset)
    el = linalg.eig(pzz)
    K = P[0,0]*H[1,0]/(H[1,0]*P[0,0]*H[1,0]+V[1,0]*R[0,0]*V[1,0]+nl[1,1])
    sh_n = sh + K*vector([z0-(ym[1]+H[1,0]*(sh[0]-sh_in[0]))])
    tf = (1-K*H[1,0])
    pn = tf*P[0,0]*tf + K*(V[1,0]*R[0,0]*V[1,0]+nl[1,1])*K
    print("Update:"+str(sh_n.n(3))+" SD:"+str(sqrt(pn).n(digits=3))+" NL:"+str(sqrt(nl[1,1]).n(3)))
    pn_root = matrix([[sqrt(pn)]])
    return(ym,ty,el,sh_n,pn_root)

for k in range(0,2) :
    st = [2*k-1]
    vt = [0.02]
    z0 = h(st,vt)
    (ym_0,ty_0,el0,sh1,pn1_root) = iterate(sh,P_root)
    (ym_1,ty_1,el1,sh2,pn2_root) = iterate(sh1,pn1_root)
    (ym_2,ty_2,el2,sh3,pn3_root) = iterate(sh2,pn2_root)
    (ym_3,ty_3,el3,sh4,pn4_root) = iterate(sh3,pn3_root)
    (ym_4,ty_4,el4,sh5,pn5_root) = iterate(sh4,pn4_root)
    (ym_5,ty_5,el5,sh6,pn6_root) = iterate(sh5,pn5_root)
    sht = sh6
    pnt_root = pn6_root
    for j in range(0,10) :
        (ym_t,ty_t,elt,sht,pnt_root) = iterate(sht,pnt_root)

    fig = plt.figure()
    fig.set_figwidth(6)
    fig.set_figheight(5)
    axs = fig.add_axes( (0.6/6,0.4/5,1-.8/6,1-.6/5))

    axs.set_xlabel("State",fontsize=12)
    axs.set_ylabel("Observations",fontsize=12)

    rands, = axs.plot( random_x,random_obs,'.',markersize=2,color='#888888')
    #rand,  = axs.plot( [sh[0]],[random_mean],'x',markersize=7,color='#333333')
    #meas, = axs.plot( [st[0]],[z0],'o',markersize=7,color='#8f00ff')
    sig0m, = axs.plot( ym_0[0],ym_0[1], 'x',markersize=7,color=clist[0])
    sig0p, = axs.plot( map(lambda u :n(u[0]),ty_0),map(lambda u:n(u[1]),ty_0),'.',markersize=6,color=clist[0])
    sig1m, = axs.plot( ym_1[0],ym_1[1], 'x',markersize=7,color=clist[1])
    sig1p, = axs.plot( map(lambda u :n(u[0]),ty_1),map(lambda u:n(u[1]),ty_1),'.',markersize=6,color=clist[1])
    sig2m, = axs.plot( ym_2[0],ym_2[1], 'x',markersize=7,color=clist[2])
    sig2p, = axs.plot( map(lambda u :n(u[0]),ty_2),map(lambda u:n(u[1]),ty_2),'.',markersize=6,color=clist[2])
    sig5m, = axs.plot( ym_5[0],ym_5[1], 'x',markersize=7,color=clist[3])
    sig5p, = axs.plot( map(lambda u :n(u[0]),ty_5),map(lambda u:n(u[1]),ty_5),'.',markersize=6,color=clist[3])
    
    e1 = Ellipse(
        xy=[ sh[0], random_mean],
        width=4*sqrt(random_xcorr_el[0][0]),
        height=4*sqrt(random_xcorr_el[0][1]),
        angle=-180/pi.n()*atan2(random_xcorr_el[1][0][1],random_xcorr_el[1][0][0]))
    e1.set_fill(False)
    e1.set_edgecolor('#333333')
    e1.set_ls('dashed')
    e1.set_lw(2)
    #axs.add_artist(e1)
    
    e2 = Ellipse(
        xy=ym_0,
        width=4*sqrt(el0[0][0]),
        height=4*sqrt(el0[0][1]),
        angle=-180/pi.n()*atan2(el0[1][0][1],el0[1][0][0]))
    e2.set_fill(False)
    e2.set_edgecolor(clist[0])
    e2.set_ls('solid')
    e2.set_lw(1.5)
    e2.set_clip_box(axs.bbox)
    axs.add_artist(e2)
    
    e3 = Ellipse(
        xy=ym_1,
        width=4*sqrt(el1[0][0]),
        height=4*sqrt(el1[0][1]),
        angle=-180/pi.n()*atan2(el1[1][0][1],el1[1][0][0]))
    e3.set_fill(False)
    e3.set_edgecolor(clist[1])
    e3.set_ls('solid')
    e3.set_lw(1.5)
    e3.set_clip_box(axs.bbox)
    axs.add_artist(e3)
    
    e4 = Ellipse(
        xy=ym_2,
        width=4*sqrt(el2[0][0]),
        height=4*sqrt(el2[0][1]),
        angle=-180/pi.n()*atan2(el2[1][0][1],el2[1][0][0]))
    e4.set_fill(False)
    e4.set_edgecolor(clist[2])
    e4.set_ls('solid')
    e4.set_lw(1.5)
    e4.set_clip_box(axs.bbox)
    axs.add_artist(e4)
    
    e5 = Ellipse(
        xy=ym_5,
        width=4*sqrt(el5[0][0]),
        height=4*sqrt(el5[0][1]),
        angle=-180/pi.n()*atan2(el5[1][0][1],el5[1][0][0]))
    e5.set_fill(False)
    e5.set_edgecolor(clist[3])
    e5.set_ls('solid')
    e5.set_lw(1.5)
    e5.set_clip_box(axs.bbox)
    axs.add_artist(e5)
    
    axs.legend( (rands,sig0p,sig1p,sig2p,sig5p), ("Random Samples","Iterate 0","Iterate 1","Iterate 2","Iterate 5"), framealpha=0.7,numpoints=1,prop={'family':'M+ 2p','weight':'ultralight','size':8})
    
    yticks = [0,2,-2,z0]
    yticks.sort()
    axs.set_yticks(yticks)
    axs.set_yticklabels(map(lambda u : u.n(digits=2),yticks), fontsize=8, family='sans-serif', weight='light')
    yticks_minor = [-3,-1,1,3]
    #axs.set_yticks(yticks_minor,minor=True)
    xticks = [ 0,-2,2,st[0]]
    xticks.sort()
    axs.set_xticks(xticks)
    xticks_minor = [-1,1]
    #axs.set_xticks(xticks_minor,minor=True)
    axs.set_xticklabels(map(lambda u : u.n(digits=2),xticks), fontsize=8, family='sans-serif', weight='light')
    axs.set_xlim([-2,2])
    axs.set_ylim([-4,2])
    axs.grid(True,'both')
    
    fig.savefig("Plots/Iterate_Example"+str(k)+".svg")
    plt.close(fig)
