from sage.symbolic.function_factory import function as fnc

#Define the numerics function we will need to evaluate stuff
BOUND = 1E-10

def sinc_f(self, x, parent=None,algorithm=None) :
    if (x<BOUND and x>-BOUND) : return 1
    else : return sin(x)/x

def cinc_f(self, x, parent=None,algorithm=None) :
    if (x<BOUND and x>-BOUND) : return 0
    else : return (cos(x)-1)/x

def dsinc_f(self, x, parent=None,algorithm=None) :
    if (x<BOUND and x>-BOUND) : return 0
    else : return ( (cos(x) - sinc_f(None,x,None,None)) / x)

def dcinc_f(self, x, parent=None,algorithm=None) :
    if (x<BOUND and x>-BOUND) : return -0.5
    else : return ( (-cinc_f(None,x,None,None) - sin(x)) / x)

#Define the symbolic versions so that this stuff makes sense
dsinc = fnc('dsinc',nargs=1,evalf_func=dsinc_f)
def diff_sinc(self, *args, **kwd) : return dsinc(args[kwd['diff_param']])
sinc = fnc('sinc',nargs=1,evalf_func=sinc_f,derivative_func=diff_sinc)

dcinc = fnc('dcinc',nargs=1,evalf_func=dcinc_f)
def diff_cinc(self, *args, **kwd) : return dcinc(args[kwd['diff_param']])
cinc = fnc('cinc',nargs=1,evalf_func=cinc_f,derivative_func=diff_cinc)


var('x y theta phi1 phi2 phid1 phid2')
state_vector = [x,y,theta,phi1,phi2,phid1,phid2]
# The state vector
# x : 0
# y : 1
# theta : 2
# phi_1 : 3
# phi_2 : 4
# phid_1 : 5
# phid_2 : 6

var('w1 w2 w3 w4 w5')
proc_noise = [w1,w2,w3,w4,w5]

var('force_v force_gam')
forcing = [force_v, force_gam]

dt = 1/2
bw = 5/4
br = 3/10

merry_1 = ( [0,-15] , 7+1/2)
merry_2 = ( [2,15], 6+1/2 )

def pedal_ahead(state,noise,forcing) : 
    vk = forcing[0] + noise[0]
    gk = forcing[1] + noise[1]
    ak = vk*dt*tan(gk)/bw
    return vector([
        state[0]+vk*dt*( sin(state[2])*cinc(ak) + cos(state[2])*sinc(ak) ) + dt*noise[2],
        state[1]+vk*dt*( sin(state[2])*sinc(ak) - cos(state[2])*cinc(ak) ) + dt*noise[3],
        state[2] + ak + dt*noise[4],
        state[3] + dt*state[5],
        state[4] + dt*state[6],
        state[5],
        state[6]
        ])

var('n1 n2')
obs_noise = [n1,n2]

def obs1(state,noise) :
    return vector([
        atan2( merry_1[0][1] + merry_1[1]*sin(state[3]) - state[1] - br*sin(state[2]), merry_1[0][0] + merry_1[1]*cos(state[3]) - state[0] - br*cos(state[2]) ) - state[2] + noise[0]
    ])

def obs2(state,noise) :
    return vector([
        atan2( merry_2[0][1] + merry_2[1]*sin(state[4]) - state[1] - br*sin(state[2]), merry_2[0][0] + merry_2[1]*cos(state[4]) - state[0] - br*cos(state[2]) ) - state[2] + noise[1]
    ])

pedal_ev = pedal_ahead(state_vector,proc_noise,forcing)

Fe = matrix([ [diff(pedal_ev[j],state_vector[k]) for k in range(0,len(state_vector))] for j in range(0,len(pedal_ev))])
Fee_s = str(Fe.list())
replace_state = [ (str(state_vector[k]),"state["+str(k)+"]") for k in range(0,len(state_vector))]
replace_proc = [ (str(proc_noise[k]),"noise["+str(k)+"]") for k in range(0,len(proc_noise))]
replace_force = [ (str(forcing[k]),"force["+str(k)+"]") for k in range(0,len(forcing))]
Fee_s = reduce(lambda u,v : u.replace(v[0],v[1]),replace_state+replace_proc+replace_force,Fee_s)
Fee_s = "matrix("+str(len(state_vector))+","+str(len(state_vector))+","+Fee_s+").n()"
Fee_e = 'def Fee(state,noise,force) : return '+Fee_s
exec(preparse(Fee_e))

Le = matrix([ [diff(pedal_ev[j],proc_noise[k]) for k in range(0,len(proc_noise))] for j in range(0,len(pedal_ev))])
Lee_s = str(Le.list())
Lee_s = reduce(lambda u,v : u.replace(v[0],v[1]),replace_state+replace_proc+replace_force,Lee_s).replace("^","**")
Lee_s = "matrix("+str(len(state_vector))+","+str(len(proc_noise))+","+Lee_s+").n()"
Lee_e = 'def Lee(state,noise,force) : return '+Lee_s
exec(preparse(Lee_e))

#These are the functions we will need to generate the C files
def show_pedal_s() : 
    for k in range(0,len(pedal_ev)) :
        print "nstate["+str(k) + "] = " + str(pedal_ev[k]).replace(str(forcing[0]),"iforce->v").replace(str(forcing[1]),"iforce->gam")

def show_Fe() :
    for k in range(0,len(pedal_ev)) :
        print str(Fe[k].list()).replace(str(forcing[0]),"iforce->v").replace(str(forcing[1]),"iforce->gam").replace("[","{").replace("]","}")

def show_Le() :
    for k in range(0,len(pedal_ev)) :
        print str(Le[k].list()).replace(str(forcing[0]),"iforce->v").replace(str(forcing[1]),"iforce->gam").replace("[","{").replace("]","}")

obs1_ev = obs1(state_vector,obs_noise)
H1e = matrix([[diff(obs1_ev[0],state_vector[k]) for k in range(0,len(state_vector))]])
replace_obs = [ (str(obs_noise[k]),"noise["+str(k)+"]") for k in range(0,len(obs_noise))]
H1ee_s = str(H1e.list())
H1ee_s = reduce(lambda u,v : u.replace(v[0],v[1]),replace_state+replace_obs,H1ee_s).replace("^","**")
H1ee_s = "matrix(1,"+str(len(state_vector))+","+H1ee_s+").n()"
H1ee_e = 'def H1ee(state,noise,force) : return '+H1ee_s
exec(preparse(H1ee_e))

V1e = matrix([[diff(obs1_ev[0],obs_noise[k]) for k in range(0,len(obs_noise))]])
V1ee_s = str(V1e.list())
V1ee_s = reduce(lambda u,v : u.replace(v[0],v[1]),replace_state+replace_obs,V1ee_s).replace("^","**")
V1ee_s = "matrix(1,"+str(len(obs_noise))+","+V1ee_s+").n()"
V1ee_e = 'def V1ee(state,noise,force) : return '+V1ee_s
exec(preparse(V1ee_e))

def show_obs1() :
    print("ang[0]="+str(obs1_ev[0]))

def show_H1e() :
    print str(H1e[0].list()).replace("[","{").replace("]","}")

def show_V1e() :
    print str(V1e[0].list()).replace("[","{").replace("]","}")

obs2_ev = obs2(state_vector,obs_noise)
H2e = matrix([[diff(obs2_ev[0],state_vector[k]) for k in range(0,len(state_vector))]])
H2ee_s = str(H2e.list())
H2ee_s = reduce(lambda u,v : u.replace(v[0],v[1]),replace_state+replace_obs,H2ee_s).replace("^","**")
H2ee_s = "matrix(1,"+str(len(state_vector))+","+H2ee_s+").n()"
H2ee_e = 'def H2ee(state,noise,force) : return '+H2ee_s
exec(preparse(H2ee_e))

V2e = matrix([[diff(obs2_ev[0],obs_noise[k]) for k in range(0,len(obs_noise))]])
V2ee_s = str(V2e.list())
V2ee_s = reduce(lambda u,v : u.replace(v[0],v[1]),replace_state+replace_obs,V2ee_s).replace("^","**")
V2ee_s = "matrix(1,"+str(len(obs_noise))+","+V2ee_s+").n()"
V2ee_e = 'def V2ee(state,noise,force) : return '+V2ee_s
exec(preparse(V2ee_e))

def show_obs2() :
    print("ang[0]="+str(obs2_ev[0]))

def show_H2e() :
    print str(H2e[0].list()).replace("[","{").replace("]","}")

def show_V2e() :
    print str(V2e[0].list()).replace("[","{").replace("]","}")

def recenter(y,cent) :
    if (y[0]-cent[0]) > pi.n() :
        y[0] = y[0]-2*pi.n()
    if (cent[0]-y[0]) > pi.n() :
        y[0] = y[0]+2*pi.n()
    return y

def obs1f(state,noise,force) :
    y = obs1(state,noise)
    return recenter(y,force)-force

def obs2f(state,noise,force) :
    y = obs2(state,noise)
    return recenter(y,force)-force
