#ifndef TABLE_H_
#define TABLE_H_

#include "linalg.h"
#include "LinearFilters.h"
#include "EKF.h"
#include "SigmaPoint.h"
#include "Smoother.h"

#include "filtersettings.h"
extern "C" {
	#include "sdf_subs.h"
}

#define MIN(X,Y) ((X) < (Y) ? (X) : (Y))

#define MAX_FILENAME 100
#define MAX_INPUT 300

struct RUN_SETTINGS {
	char basename[MAX_FILENAME];
	char runname[MAX_FILENAME];
	int nruns;
	LINEAR_FILTER *filters;
	int nfilter;
};

struct TRICYCLE_FORCING {
	NT v;
	NT gam;
};

struct MERRY_GO_ROUND {
	NT rho;
	NT mx,my;
};

struct MERRY_FORCING {
	int ndx;
	vec<1> *center;
};


int read_in_settings(const char *,RUN_SETTINGS &);
int read_in_scenario(char*, int, GAUSSIAN<7>&, vec<7> &, vec<2>**, vec<5>**, GAUSSIAN<5>&, vec<2>**, GAUSSIAN<2>&, vec<2>**);

void runFilter(int,int,GAUSSIAN<7>&,vec<7>*,vec<5>*,vec<2>*,MERRY_FORCING*,TRICYCLE_FORCING*,LINEAR_FILTER&);

vec<7> moveForward(mat<7,7> *, mat<7,5> *, const vec<7> &, const vec<5> &, TRICYCLE_FORCING *);

void runEKF(int,GAUSSIAN<7>&,TRICYCLE_FORCING*,MERRY_FORCING*,GAUSSIAN<7>*,GAUSSIAN<7>*,ADVANCEMENT<7,5>*,MEASUREMENT<7,1,2>*,vec<7>*,vec<5>*,vec<2>*);
void runBLSF(int,GAUSSIAN<7>&,TRICYCLE_FORCING*,MERRY_FORCING*,GAUSSIAN<7>*,GAUSSIAN<7>*,ADVANCEMENT<7,5>*,MEASUREMENT<7,1,2>*,vec<7>*,vec<5>*,vec<2>*);
void runBackEKF(int,GAUSSIAN<7>*,GAUSSIAN<7>*,ADVANCEMENT<7,5>*,vec<7>*,vec<5>*,MEASUREMENT<7,1,2>*,vec<2>*);

void runMerg1(SIGMAPOINT_SETTINGS<7+5> &, NT, int, GAUSSIAN<7> &, TRICYCLE_FORCING *, MERRY_FORCING *, GAUSSIAN<7> *, GAUSSIAN<7> *, ADVANCEMENT<7,5> *, MEASUREMENT<7,1,2> *, GAUSSIAN<7>*);
void runMerg2(SIGMAPOINT_SETTINGS<7+2> &, NT, int , GAUSSIAN<7> &, TRICYCLE_FORCING *, MERRY_FORCING *, GAUSSIAN<7> *, GAUSSIAN<7> *, ADVANCEMENT<7,5> *, MEASUREMENT<7,1,2> *, GAUSSIAN<7>*);
void runSigma(SIGMAPOINT_SETTINGS<7+5> &, SIGMAPOINT_SETTINGS<7+2> &, NT, NT, int , GAUSSIAN<7> &, TRICYCLE_FORCING *, MERRY_FORCING *, GAUSSIAN<7> *, GAUSSIAN<7> *, ADVANCEMENT<7,5> *, MEASUREMENT<7,1,2> *, GAUSSIAN<7>*);
//void runSigmaTR(SIGMAPOINT_SETTINGS &, int , GAUSSIAN<7> &, TRICYCLE_FORCING *, MERRY_FORCING *, GAUSSIAN<7> *, GAUSSIAN<7> *, ADVANCEMENT<7,5> *, MEASUREMENT<7,1> *, GAUSSIAN<7>*);
void runBackSigma(int, GAUSSIAN<7>*, GAUSSIAN<7>*, ADVANCEMENT<7,5>*, GAUSSIAN<7>*, vec<5>*, MEASUREMENT<7,1,2>*, vec<2>*);

void clean_datafiles(void);
void add_datafile(char*, char, int, int, pos*, void*, int);

#endif /* TABLE_H_ */
