#include "Space.h"
#include <math.h>

// State Vector
// Pos 1-3
// Vel 4-6
// Q 7-10
// Rvel 11-13
// ab 14-16
// gb 17-19

#define BOUND 1E-10
NT cinc(NT ang){
	if (ang<BOUND && ang>-BOUND)
		return 0;
	return (cos(ang)-1)/ang;
}

NT sinc(NT ang){
	if (ang<BOUND && ang>-BOUND)
		return 1;
	return sin(ang)/ang;
}

NT dcinc(NT ang){
	if (ang<BOUND && ang>-BOUND)
		return -0.5;
	return ( (-cinc(ang) - sin(ang))/ang );
}
NT dsinc(NT ang){
	if (ang<BOUND && ang>-BOUND)
		return 0;
	return ( (cos(ang)-sinc(ang))/ang);
}

#define SQ(x) ((x)*(x))

vec<7> moveForward(mat<7,7> *F, mat<7,5> *L, const vec<7> &state, const vec<5> &w, TRICYCLE_FORCING *iforce) {
	vec<7> nstate;

	NT x,y,th,phi1,phi2,phid1,phid2;
	NT w1,w2,w3,w4,w5;

	NT vM,gammaM,ug;

	x 		= state[0];
	y 		= state[1];
	th 		= state[2];
	phi1 	= state[3];
	phi2	= state[4];
	phid1	= state[5];
	phid2	= state[6];

	w1		= w[0];
	w2		= w[1];
	w3		= w[2];
	w4		= w[3];
	w5		= w[4];

	vM = iforce->v + w1;
	gammaM = iforce->gam + w2;
	ug = dt * vM * tan(gammaM) / bw;
	nstate[0] = x + vM*dt*( sin(th)*cinc(ug) + cos(th)*sinc(ug) ) + dt*w3;
	nstate[1] = y + vM*dt*(-cos(th)*cinc(ug) + sin(th)*sinc(ug) ) + dt*w4;
	nstate[2] = th + ug + dt*w5;
	nstate[3] = phi1 + dt*phid1;
	nstate[4] = phi2 + dt*phid2;
	nstate[5] = phid1;
	nstate[6] = phid2;

	if (F != NULL) {
		NT tds[7][7] = {
				{1, 0, vM*dt*(cos(th)*cinc(ug)-sin(th)*sinc(ug)), 0, 0, 0, 0},
				{0, 1, vM*dt*(sin(th)*cinc(ug)+cos(th)*sinc(ug)), 0, 0, 0, 0},
				{0, 0, 1, 0, 0, 0, 0},
				{0, 0, 0, 1, 0, dt, 0},
				{0, 0, 0, 0, 1, 0, dt},
				{0, 0, 0, 0, 0, 1, 0},
				{0, 0, 0, 0, 0, 0, 1}
		};
		*F = tds;
	}
	if (L != NULL ) {
		NT tdw[7][5] = {
				{dt*(sin(th)*cinc(ug)+cos(th)*sinc(ug))+vM*dt*(sin(th)*dcinc(ug)*dt*tan(gammaM)/bw+cos(th)*dsinc(ug)*dt*tan(gammaM)/bw),
						vM*dt*(sin(th)*dcinc(ug)*dt*vM*(1+tan(gammaM)*tan(gammaM))/bw+cos(th)*dsinc(ug)*dt*vM*(1+tan(gammaM)*tan(gammaM))/bw),
						dt,0,0},
				{dt*(-cos(th)*cinc(ug)+sin(th)*sinc(ug))+vM*dt*(-cos(th)*dcinc(ug)*dt*tan(gammaM)/bw+sin(th)*dsinc(ug)*dt*tan(gammaM)/bw),
						vM*dt*(-cos(th)*dcinc(ug)*dt*vM*(1+tan(gammaM)*tan(gammaM))/bw+sin(th)*dsinc(ug)*dt*vM*(1+tan(gammaM)*tan(gammaM))/bw),
						0,dt,0},
				{dt*tan(gammaM)/bw,dt*vM*(1+tan(gammaM)*tan(gammaM))/bw,0,0,dt},
				{0, 0, 0, 0, 0},
				{0, 0, 0, 0, 0},
				{0, 0, 0, 0, 0},
				{0, 0, 0, 0, 0}
		};
		*L = tdw;
	}
	return nstate;
}

vec<1> observer1(mat<1,7> *H, mat<1,2> *V, const vec<7> &state, const vec<2> &n, MERRY_FORCING *id){
	vec<1> ang;
	NT x,y,th;
	NT mx,my,rho;

	x 		= state[0];
	y 		= state[1];
	th 		= state[2];

	mx = merrys[0].mx;
	my = merrys[0].my;
	rho = merrys[0].rho;

	NT phi = state[3];

	ang[0] = atan2((my + rho*sin(phi)-y-br*sin(th)),
			(mx+rho*cos(phi)-x-br*cos(th)))
			-th+n[0];

	if (H != NULL) {
		NT tH[7] = {
			(my+rho*sin(phi)-y-br*sin(th))/SQ(mx+rho*cos(phi)-x-br*cos(th))/(1+SQ(my+rho*sin(phi)-y-br*sin(th))/SQ(mx+rho*cos(phi)-x-br*cos(th))),
			-1/(mx+rho*cos(phi)-x-br*cos(th))/(1+SQ(my+rho*sin(phi)-y-br*sin(th))/SQ(mx+rho*cos(phi)-x-br*cos(th))),
			(-br*cos(th)/(mx+rho*cos(phi)-x-br*cos(th))-(my+rho*sin(phi)-y-br*sin(th))/SQ(mx+rho*cos(phi)-x-br*cos(th))*br*sin(th))/(1+SQ(my+rho*sin(phi)-y-br*sin(th))/SQ(mx+rho*cos(phi)-x-br*cos(th)))-1,
			(rho*cos(phi)/(mx+rho*cos(phi)-x-br*cos(th))+(my+rho*sin(phi)-y-br*sin(th))/SQ(mx+rho*cos(phi)-x-br*cos(th))*rho*sin(phi))/(1+SQ(my+rho*sin(phi)-y-br*sin(th))/SQ(mx+rho*cos(phi)-x-br*cos(th))),
			0,
			0,
			0};
		*H = tH;
	}
	if (V != NULL){
		NT tV[2] = {1,0};
		*V = tV;
	}

	if(id->center != NULL) {
		if ( (ang[0] - (*(id->center))[0] ) > 3.14159)
			ang[0] = ang[0] - (2*3.14159);
		else if ( ((*(id->center))[0] - ang[0]) > 3.14159)
			ang[0] = ang[0] + (2*3.14159);
		ang = ang - (*(id->center));
	}

	return ang;
}

vec<1> observer2(mat<1,7> *H, mat<1,2> *V, const vec<7> &state, const vec<2> &n, MERRY_FORCING *id){
	vec<1> ang;
	NT x,y,th;
	NT mx,my,rho;

	x 		= state[0];
	y 		= state[1];
	th 		= state[2];

	mx = merrys[1].mx;
	my = merrys[1].my;
	rho = merrys[1].rho;

	NT phi = state[3+1];

	ang[0] = atan2((my + rho*sin(phi)-y-br*sin(th)),
			(mx+rho*cos(phi)-x-br*cos(th)))
			-th+n[1];

	if (H != NULL) {
		NT tH[7] = {(my+rho*sin(phi)-y-br*sin(th))/SQ(mx+rho*cos(phi)-x-br*cos(th))/(1+SQ(my+rho*sin(phi)-y-br*sin(th))/SQ(mx+rho*cos(phi)-x-br*cos(th))),
			-1/(mx+rho*cos(phi)-x-br*cos(th))/(1+SQ(my+rho*sin(phi)-y-br*sin(th))/SQ(mx+rho*cos(phi)-x-br*cos(th))),
			(-br*cos(th)/(mx+rho*cos(phi)-x-br*cos(th))-(my+rho*sin(phi)-y-br*sin(th))/SQ(mx+rho*cos(phi)-x-br*cos(th))*br*sin(th))/(1+SQ(my+rho*sin(phi)-y-br*sin(th))/SQ(mx+rho*cos(phi)-x-br*cos(th)))-1,
			0,
			((rho*cos(phi)/(mx+rho*cos(phi)-x-br*cos(th))+(my+rho*sin(phi)-y-br*sin(th))/SQ(mx+rho*cos(phi)-x-br*cos(th))*rho*sin(phi))/(1+SQ(my+rho*sin(phi)-y-br*sin(th))/SQ(mx+rho*cos(phi)-x-br*cos(th)))),
			0,
			0};
		*H = tH;
	}
	if (V != NULL){
		NT tV[2] = {0,1};
		*V = tV;
	}

	if(id->center != NULL) {
		if ( (ang[0] - (*(id->center))[0] ) > 3.14159)
			ang[0] = ang[0] - 3.14159*2;
		else if ( ((*(id->center))[0] - ang[0]) > 3.14159)
			ang[0] = ang[0] + 3.14159*2;
		ang = ang - (*(id->center));
	}

	return ang;
}
